//
// tournament - creation of playing schedules for tournaments
// Copyright (C) 2011-2022  Joachim Reichel <joachim.reichel@posteo.de>
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this program; if not, write to the Free Software Foundation, Inc.,
// 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
//

/// \file
/// \brief Main file.

#include <cassert>
#include <csignal>
#include <cmath>
#include <cstdlib>
#include <ctime>
#include <iostream>
#include <limits>
#include <list>
#include <memory>
#include <numeric>
#include <sstream>
#include <vector>

#ifdef HAVE_OPENMP
#include <omp.h>
#endif // HAVE_OPENMP

#include "timer.h"

/// Number of male players
int K;
/// Number of female players
int L;
/// Total number of players (K plus L)
int KpL;
/// Number of rounds
int M;
/// Number of courts
int N;
/// Number of courts in the last round
int last_N;
/// Number of matches (M-1 times N plus last_N)
int MtN;
/// Volatility of the simulated annealing
double V;
/// Lower bound of a tournament
int g_lower_bound;

/// Flag to interrupt computation, e.g., in case of Ctrl-C.
bool g_interrupt = false;

/// Helper class that supports output with padding.
///
/// Its output operator invokes the output operator of the base class of type T.
/// If necessary, the output is then padded to the given width.
template<class T>
class Padding : public T
{
public:
    /// Constructor.
    ///
    /// Wraps the instance \p t such that its output operator writes at least
    /// \p width characters.
    Padding( const T& t, size_t width): T( t), m_width( width) { };

    /// The desired width.
    const size_t m_width;
};

/// Output operator
template<class T>
std::ostream& operator<<( std::ostream& stream, const Padding<T>& p)
{
    std::ostringstream s;
    s << static_cast<const T&>( p);
    for( size_t i = s.str().size(); i < p.m_width; ++i)
        s << " ";
    stream << s.str();
    return stream;
}

/// Represents a player.
///
/// Players are identified by an ID. The IDs 0..#K-1 indicate the male players,
/// the IDs #K..#K+#L-1 indicate the female players. The output operator uses
/// letters 'A', 'B', ... for male players, and numbers 0, 1, ... for female
/// players.
class Player
{
public:
    /// Constructor.
    ///
    /// Constructs a player with ID 0.
    Player() = default;

    /// Constructor.
    ///
    /// Constructs a player with ID \p id.
    explicit Player( int id) : m_id( id) { }

    /// Sets the player ID.
    ///
    /// \param id   The new player ID.
    void set_id( int id) { assert( id >= 0); assert( id < KpL); m_id = id; }

    /// Returns the player ID.
    int get_id() const { return m_id; }

    /// Indicates whether the player is male.
    bool is_male() const { return m_id < K; }

    /// Indicates whether the players is female.
    bool is_female() const { return m_id >= K; }

    /// Equality operator.
    bool operator==( const Player& rhs) const { return m_id == rhs.m_id; }

    /// Inequality operator.
    bool operator!=( const Player& rhs) const { return m_id != rhs.m_id; }

private:
    /// The player ID.
    int m_id = 0;
};

/// Output operator
std::ostream& operator<<( std::ostream& stream, const Player& p)
{
    int id = p.get_id();
    if( p.is_male())
        stream << static_cast<char>( 'A' + id);
    else
        stream << id - (K-1);
    return stream;
}

/// Represents a match.
///
/// A match consists of four players, identified by indices 0 to 3. The first
/// team consists of players 0 and 1, the second team consists of players 2 and
/// 3. The players 0 and 2 are male, the players 1 and 3 are female.
class Match
{
public:
    /// Constructor.
    ///
    /// Constructs a match with players 0 and K against players 1 and K+1.
    Match()
    {
        m_player[0] = Player( 0);
        m_player[1] = Player( K);
        m_player[2] = Player( 1);
        m_player[3] = Player( K+1);
    }

    /// Sets a player of the match.
    ///
    /// \param index   The player to change.
    /// \param player  The new player.
    void set_player( int index, const Player& player)
    {
        assert( index >= 0);
        assert( index < 4);
        m_player[index] = player;
    }

    /// Returns a player of the match
    ///
    /// \param index   The player to return.
    /// \return        The player with index \p index of the match.
    const Player& get_player( int index) const
    {
        assert( index >= 0);
        assert( index < 4);
        return m_player[index];
    }

    /// Indicates whether the match is valid.
    ///
    /// \return   \c true if the male players are different and the female
    ///           players are different
    bool is_valid() const
    {
        assert( m_player[0].is_male());
        assert( L > 0 ? m_player[1].is_female() : m_player[1].is_male());
        assert( m_player[2].is_male());
        assert( L > 0 ? m_player[3].is_female() : m_player[3].is_male());

        if( L > 0)
            return (m_player[0] != m_player[2]) && (m_player[1] != m_player[3]);

        return (m_player[0] != m_player[1]) && (m_player[0] != m_player[2])
            && (m_player[0] != m_player[3]) && (m_player[1] != m_player[2])
            && (m_player[1] != m_player[3]) && (m_player[2] != m_player[3]);
    }

private:
    /// The players of the match.
    Player m_player[4];
};

/// Output operator
std::ostream& operator<<( std::ostream& stream, const Match& m)
{
    stream << m.get_player( 0) << Padding<Player>( m.get_player( 1), L>9 ? 2 : 1) << " : "
           << m.get_player( 2) << Padding<Player>( m.get_player( 3), L>9 ? 2 : 1);
    return stream;
}

/// Represents a tournament.
///
/// A tournament consists of rounds. In each round there is exactly one match
/// per court.
///
/// Uses the global variables \p M and \p N.
class Tournament
{
public:
    /// Constructor.
    ///
    /// Creates a tournament with #M rounds and #N courts.
    Tournament()
      : m_matches( std::make_unique<Match[]>( MtN))
    {
        int k = 0;
        int l = 0;
        for( int i = 0; i < MtN; ++i) {
            m_matches[i].set_player( 0, Player( 0 + k++ % K));
            m_matches[i].set_player( 1, L == 0 ? Player( 0 + k++ % K) : Player( K + l++ % L));
            m_matches[i].set_player( 2, Player( 0 + k++ % K));
            m_matches[i].set_player( 3, L == 0 ? Player( 0 + k++ % K) : Player( K + l++ % L));
        }
    }

    /// Copy constructor.
    Tournament( const Tournament& other)
      : m_matches( std::make_unique<Match[]>( MtN))
    {
        std::copy( other.m_matches.get(), other.m_matches.get() + MtN, m_matches.get());
    }

    /// Move constructor.
    Tournament( Tournament&& other) = default;

    /// Assignment operator.
    Tournament& operator=( const Tournament& rhs)
    {
        if( this == &rhs)
            return *this;

        std::copy( rhs.m_matches.get(), rhs.m_matches.get() + MtN, m_matches.get());
        return *this;
    }

    /// Move assignment operator.
    Tournament& operator=( Tournament&& rhs) = default;

    /// Destructor.
    ~Tournament() = default;

    /// Returns the number of rounds.
    int get_rounds() const { return M; }

    /// Returns the number of courts.
    int get_courts( int round) const { return round < M-1 ? N : last_N; }

    /// Returns a match.
    ///
    /// \param round   The round of the match to return.
    /// \param court   The court of the match to return.
    /// \return        The match on court \p court in round \p round.
    const Match& get_match( int round, int court) const
    {
        return m_matches[ round * N + court];
    }

    /// Returns a match.
    ///
    /// \param round   The round of the match to return.
    /// \param court   The court of the match to return.
    /// \return        The match on court \p court in round \p round.
    Match& get_match( int round, int court)
    {
        return m_matches[ round * N + court];
    }

    /// Returns a match.
    ///
    /// \param index   The index of the match to return.
    /// \return        The index-th match.
    const Match& get_match( int index) const
    {
        return m_matches[ index];
    }

    /// Returns a match.
    ///
    /// \param index   The index of the match to return.
    /// \return        The index-th match.
    Match& get_match( int index)
    {
        return m_matches[ index];
    }

    /// Returns the round corresponding to the given match index.
    int get_round( int index)
    {
        return index / N;
    }

    /// Returns the court corresponding to the given match index.
    int get_court( int index)
    {
        return index % N;
    }

private:
    /// The matches.
    std::unique_ptr<Match[]> m_matches;
};

/// Output operator
std::ostream& operator<<( std::ostream& stream, const Tournament& t)
{
    stream << "          ";
    if( M > 9) stream << " ";
    if( L > 9) stream << " ";
    for( int j = 0; j < t.get_courts( 0); ++j) {
         stream << "court " << j+1 << "   ";
         if( L > 9) stream << "  ";
    }
    stream << std::endl;

    for( int i = 0; i < t.get_rounds(); ++i) {
        stream << "round " << i+1 << "   ";
        if( M > 9 && i < 9) stream << " ";
        for( int j = 0; j < t.get_courts( i); ++j)
            stream << t.get_match( i, j) << "   ";
        stream << std::endl;
    }
    return stream;
}

/// Compute penalties for every player that plays multiple times in the same
/// round.
///
/// \param t        The tournament for which to compute the penalty.
/// \param weight   The weight for this penalty.
/// \param buffer   A buffer for temporary computations of size #KpL * #KpL
/// \param stream   If not \c NULL, dump a message for each penalty to this stream.
/// \return         The computed penalty.
int penalty_multiple_per_round( const Tournament& t, int weight, char* buffer, std::ostream* stream)
{
    int total_penalty = 0;
    bool dumped = false;

    for( int i = 0; i < t.get_rounds(); ++i) {

        for( int p = 0; p < KpL; ++p)
            buffer[p] = 0;

        // For all p in [0,KpL): buffer[p] indicates how often player p
        // plays in round i.
        for( int j = 0; j < t.get_courts( i); ++j) {
            const Match& match = t.get_match( i, j);
            buffer[ match.get_player( 0).get_id() ]++;
            buffer[ match.get_player( 1).get_id() ]++;
            buffer[ match.get_player( 2).get_id() ]++;
            buffer[ match.get_player( 3).get_id() ]++;
        }

        // Let c indicate how ofter player p plays in round i.
        // If c > 1 add penalty c-1.
        for( int p = 0; p < KpL; ++p) {
            int c = buffer[p];
            if( c <= 1)
                 continue;
            int penalty = weight * (c-1);
            total_penalty += penalty;
            if( stream) {
                *stream << "player " << Player( p) << " plays "
                        << c << " times in round " << i+1
                        << " (penalty: " << penalty << ")" << std::endl;
                dumped = true;
            }
        }
    }

    if( dumped) {
        *stream << "penalties for multiple games per round: " << total_penalty << std::endl;
        *stream << std::endl;
    }
    return total_penalty;
}

/// Compute penalties for every pair of players that play together multiple times.
///
/// \param t        The tournament for which to compute the penalty.
/// \param weight   The weight for this penalty.
/// \param buffer   A buffer for temporary computations of size #KpL * #KpL
/// \param stream   If not \c NULL, dump a message for each penalty to this stream.
/// \return         The computed penalty.
int penalty_same_teammate( const Tournament& t, int weight, char* buffer, std::ostream* stream)
{
    int total_penalty = 0;
    bool dumped = false;
    for( int p = 0; p < KpL*KpL; ++p)
        buffer[p] = 0;

    // For all p < q in [0,KpL): buffer[ p*KpL + q ] indicates how often
    // player p and q play together.
    for( int i = 0; i < MtN; ++i) {
        const Match& match = t.get_match( i);
        int p0 = match.get_player( 0).get_id();
        int p1 = match.get_player( 1).get_id();
        int p2 = match.get_player( 2).get_id();
        int p3 = match.get_player( 3).get_id();
        buffer[ p0 <= p1 ? p0*KpL + p1 : p1*KpL + p0 ]++;
        buffer[ p2 <= p3 ? p2*KpL + p3 : p3*KpL + p2 ]++;
    }

    // Let c indicate how ofter player p and q play together.
    // If c > 1 add penalty (c-1) * (c-1).
    for( int p = 0; p < KpL; ++p) {
        for( int q = p+1; q < KpL; ++q) {
            int c = buffer[ p*KpL + q ];
            if( c <= 1)
                continue;
            int penalty = weight * (c-1) * (c-1);
            total_penalty += penalty;
            if( stream) {
                *stream << "players " << Player( p) << " and " << Player( q)
                        << " play " << c << " times as a team"
                        << " (penalty: " << penalty << ")" << std::endl;
                dumped = true;
            }
        }
    }

    assert( total_penalty % weight == 0);

    if( dumped) {
        *stream << "penalties for same teammate: " << total_penalty << std::endl;
        *stream << std::endl;
    }
    return total_penalty;
}

/// Compute penalties for every pair of players that play against each other
/// multiple times.
///
/// \param t        The tournament for which to compute the penalty.
/// \param weight   The weight for this penalty.
/// \param buffer   A buffer for temporary computations of size #KpL * #KpL
/// \param stream   If not \c NULL, dump a message for each penalty to this stream.
/// \return         The computed penalty.
int penalty_same_opponent( const Tournament& t, int weight, char* buffer, std::ostream* stream)
{
    int total_penalty = 0;
    bool dumped = false;
    for( int p = 0; p < KpL*KpL; ++p)
        buffer[p] = 0;

    // For all p < q in [0,KpL): buffer[ p*KpL + q ] indicates how often
    // player p and q play against each other.
    for( int i = 0; i < MtN; ++i) {
        const Match& match = t.get_match( i);
        int p0 = match.get_player( 0).get_id();
        int p1 = match.get_player( 1).get_id();
        int p2 = match.get_player( 2).get_id();
        int p3 = match.get_player( 3).get_id();
        buffer[ p0 <= p2 ? p0*KpL + p2 : p2*KpL + p0 ]++;
        buffer[ p0 <= p3 ? p0*KpL + p3 : p3*KpL + p0 ]++;
        buffer[ p1 <= p2 ? p1*KpL + p2 : p2*KpL + p1 ]++;
        buffer[ p1 <= p3 ? p1*KpL + p3 : p3*KpL + p1 ]++;
    }

    // Let c indicate how ofter player p and q play against each other.
    // If c > 1 add penalty (c-1) * (c-1).
    for( int p = 0; p < KpL; ++p) {
        for( int q = p+1; q < KpL; ++q) {
            int c = buffer[ p*KpL + q ];
            if( c <= 1)
                continue;
            int penalty = weight * (c-1) * (c-1);
            total_penalty += penalty;
            if( stream) {
                *stream << "players " << Player( p) << " and " << Player( q)
                        << " play " << c << " times against each other"
                        << " (penalty: " << penalty << ")" << std::endl;
                dumped = true;
            }
        }
    }

    assert( total_penalty % weight == 0);

    if( dumped) {
        *stream << "penalties for same opponent: " << total_penalty << std::endl;
        *stream << std::endl;
    }
    return total_penalty;
}

/// Compute penalties for subsequent matches without break.
///
/// \param t        The tournament for which to compute the penalty.
/// \param weight   The weight for this penalty.
/// \param buffer   A buffer for temporary computations of size #KpL * #KpL
/// \param stream   If not \c NULL, dump a message for each penalty to this stream.
/// \return         The computed penalty.
int penalty_subsequent_matches( const Tournament& t, int weight, char* buffer, std::ostream* stream)
{
    int total_penalty = 0;
    bool dumped = false;
    for( int p = 0; p < KpL; ++p)
        buffer[ p ] = 0;
    for( int p = 0; p < KpL; ++p)
        buffer[ KpL + p ] = -1;

    // For all p in [0,KpL): buffer[ p ] indicates the maximum number of subsequent
    // matches of player p (so far), buffer[ KpL + p ] indicates the last round
    // where player p had a break, and buffer [ 2*KpL + p ] indicates whether
    // player p has a match in this round.
    for( int i = 0; i < t.get_rounds(); ++i) {
        for( int p = 0; p < KpL; ++p)
            buffer[ 2*KpL + p ] = 0;
        for( int j = 0; j < t.get_courts( i); ++j) {
            const Match& match = t.get_match( i, j);
            for( int k = 0; k < 4; ++k) {
                int p = match.get_player( k).get_id();
                buffer[ 2*KpL + p ] = 1;
            }
        }
        for( int p = 0; p < KpL; ++p)
            if( buffer[ 2*KpL + p ] == 1) {
                int d = i - buffer[ KpL + p ];
                if( d > buffer[ p ])
                    buffer[ p ] = static_cast<char>( d);
            } else
                buffer[ KpL + p ] = static_cast<char>( i);
    }

    // Let c indicate the maximum number of subsequent matches of player p.
    // If c > 1 add penalty (c-1) * (c-1).
    for( int p = 0; p < KpL; ++p) {
        int c = buffer[ p ];
        if( c <= 1)
            continue;
        int penalty = weight * (c-1) * (c-1);
        total_penalty += penalty;
        if( stream) {
            *stream << "player " << Player( p) << " has "
                    << c << " subsequent matches"
                    << " (penalty: " << penalty << ")" << std::endl;
            dumped = true;
        }
    }

    if( dumped) {
        *stream << "penalties for subsequent matches: " << total_penalty << std::endl;
        *stream << std::endl;
    }
    return total_penalty;
}

/// Compute penalties for subsequent breaks.
///
/// \param t        The tournament for which to compute the penalty.
/// \param weight   The weight for this penalty.
/// \param buffer   A buffer for temporary computations of size #KpL * #KpL
/// \param stream   If not \c NULL, dump a message for each penalty to this stream.
/// \return         The computed penalty.
int penalty_subsequent_breaks( const Tournament& t, int weight, char* buffer, std::ostream* stream)
{
    int total_penalty = 0;
    bool dumped = false;
    for( int p = 0; p < KpL; ++p)
        buffer[ p ] = 0;
    for( int p = 0; p < KpL; ++p)
        buffer[ KpL + p ] = -1;

    // For all p in [0,KpL): buffer[ p ] indicates the maximum number of subsequent
    // breaks of player p (so far), buffer[ KpL + p ] indicates the last round
    // where player p had a match.
    for( int i = 0; i < t.get_rounds(); ++i) {
        for( int j = 0; j < t.get_courts( i); ++j) {
            const Match& match = t.get_match( i, j);
            for( int k = 0; k < 4; ++k) {
                int p = match.get_player( k).get_id();
                int d = i - buffer[ KpL + p ] - 1;
                if( d > buffer[ p ])
                    buffer[ p ] = static_cast<char>( d);
                buffer[ KpL + p ] = static_cast<char>( i);
            }
        }
    }
    for( int p = 0; p < KpL; ++p) {
        int d = t.get_rounds() - buffer[ KpL + p ] - 1;
        if( d > buffer[ p ])
            buffer[ p ] = static_cast<char>( d);
    }

    // Let c indicate the maximum number of subsequent breaks of player p.
    // If c > 1 add penalty (c-1) * (c-1).
    for( int p = 0; p < KpL; ++p) {
        int c = buffer[ p ];
        if( c <= 1)
            continue;
        int penalty = weight * (c-1) * (c-1);
        total_penalty += penalty;
        if( stream) {
            *stream << "player " << Player( p) << " has "
                    << c << " subsequent breaks"
                    << " (penalty: " << penalty << ")" << std::endl;
            dumped = true;
        }
    }

    if( dumped) {
        *stream << "penalties for subsequent breaks: " << total_penalty << std::endl;
        *stream << std::endl;
    }
    return total_penalty;
}

/// Evaluate a tournament.
///
/// The value of a tournament is given by the weighted penalties for different
/// constraints.
///
/// \param t        The tournament to evaluate.
/// \param w0       The weight for #penalty_multiple_per_round().
/// \param w1       The weight for #penalty_same_teammate().
/// \param w2       The weight for #penalty_same_opponent().
/// \param w3       The weight for #penalty_subsequent_matches().
/// \param w4       The weight for #penalty_subsequent_breaks().
/// \param buffer   A buffer for temporary computations of size #KpL * #KpL
/// \param stream   If not \c NULL, dump a message for each penalty to this stream.
/// \return         The computed value.
int evaluate( const Tournament& t, int w0, int w1, int w2, int w3, int w4, char* buffer, std::ostream* stream)
{
    return (w0 == 0 ? 0 : penalty_multiple_per_round( t, w0, buffer, stream))
         + (w1 == 0 ? 0 : penalty_same_teammate( t, w1, buffer, stream))
         + (w2 == 0 ? 0 : penalty_same_opponent( t, w2, buffer, stream))
         + (w3 == 0 ? 0 : penalty_subsequent_matches( t, w3, buffer, stream))
         + (w4 == 0 ? 0 : penalty_subsequent_breaks( t, w4, buffer, stream));
}

/// Perform a random mutation on a tournament.
///
/// The random mutation consists of a swap of two randomly picked players. Both
/// players are either male of female.
///
/// \param t   The tournament to mutate.
/// \param i1  Used to store the mutation for later undo.
/// \param i2  Used to store the mutation for later undo.
/// \param p1  Used to store the mutation for later undo.
/// \param p2  Used to store the mutation for later undo.
void mutation( Tournament& t, int& i1, int& i2, int& p1, int& p2)
{
    bool valid_and_meaningful;

    do {
        i1 = rand() % MtN;
        Match& m1 = t.get_match( i1);

        i2 = rand() % MtN;
        Match& m2 = t.get_match( i2);

        int r = rand();
        if( L == 0) {
            p1 =  r       & 3;
            p2 = (r >> 2) & 3;
        } else {
            p1 = ((r & 4) >> 1) + (r & 1);
            p2 = ((r & 2)     ) + (r & 1);
        }

        valid_and_meaningful = L == 0

            ?    m1.get_player( p1) != m2.get_player( 0)
              && m1.get_player( p1) != m2.get_player( 1)
              && m1.get_player( p1) != m2.get_player( 2)
              && m1.get_player( p1) != m2.get_player( 3)
              && m2.get_player( p2) != m1.get_player( 0)
              && m2.get_player( p2) != m1.get_player( 1)
              && m2.get_player( p2) != m1.get_player( 2)
              && m2.get_player( p2) != m1.get_player( 3)

            :    m1.get_player( p1) != m2.get_player( p2)
              && m1.get_player( p1) != m2.get_player( p2 ^ 2)
              && m2.get_player( p2) != m1.get_player( p1 ^ 2);
    } while( !valid_and_meaningful);

    Match& m1 = t.get_match( i1);
    Match& m2 = t.get_match( i2);

    Player p = m1.get_player( p1);
    m1.set_player( p1, m2.get_player( p2));
    m2.set_player( p2, p);

    assert( m1.is_valid() && m2.is_valid());
}

/// Perform a random mutation on a tournament.
///
/// The random mutation consists of a swap of two randomly picked matches.
///
/// \param t   The tournament to mutate.
/// \param i1  Used to store the mutation for later undo.
/// \param i2  Used to store the mutation for later undo.
void mutation_match( Tournament& t, int& i1, int& i2)
{
    i1 = rand() % MtN;
    int r1 = t.get_round( i1);

    int r2;
    do {
        i2 = rand() % MtN;
        r2 = t.get_round( i2);
    } while( N > 1 && r2 == r1);

    Match& m1 = t.get_match( i1);
    Match& m2 = t.get_match( i2);

    Match tmp = m1;
    m1 = m2;
    m2 = tmp;
}

/// Reverts a given mutation on a tournament.
///
/// \param t   The tournament to modify.
/// \param i1  The mutation to undo.
/// \param i2  The mutation to undo.
/// \param p1  The mutation to undo.
/// \param p2  The mutation to undo.
void undo_mutation( Tournament& t, int i1, int i2, int p1, int p2)
{
    Match& m1 = t.get_match( i1);
    Match& m2 = t.get_match( i2);

    Player p = m1.get_player( p1);
    m1.set_player( p1, m2.get_player( p2));
    m2.set_player( p2, p);
}

/// Reverts a given mutation on a tournament.
///
/// \param t   The tournament to modify.
/// \param i1  The mutation to undo.
/// \param i2  The mutation to undo.
void undo_mutation_match( Tournament& t, int i1, int i2)
{
    Match& m1 = t.get_match( i1);
    Match& m2 = t.get_match( i2);

    Match tmp = m1;
    m1 = m2;
    m2 = tmp;
}

int run( int w0, int w1, int w2, int w3, int w4, int max_iteration, char* buffer, Tournament& best_tournament)
{
    Tournament tournament = best_tournament;
    int i = 0;
    int best_iteration = 0;
    int value = evaluate( tournament, w0, w1, w2, w3, w4, buffer, nullptr);
    int best_value = value;
    int i1, i2, p1, p2;

    while( i < max_iteration && !g_interrupt) {

        w1 + w2 > 0 ? mutation( tournament, i1, i2, p1, p2) : mutation_match( tournament, i1, i2);
        int new_value = evaluate( tournament, w0, w1, w2, w3, w4, buffer, nullptr);

        int delta = new_value - value;
        if( delta >= 0) {
#ifdef _WIN32
            double rnd = static_cast<double>( rand()) / RAND_MAX;
#else // _WIN32
            double rnd = drand48();
#endif // _WIN32
            if( rnd <= exp( -V * delta / value))
                value = new_value;
            else
                w1 + w2 > 0 ? undo_mutation( tournament, i1, i2, p1, p2) : undo_mutation_match( tournament, i1, i2);
        } else {
            value = new_value;
            if( value < best_value) {
                best_tournament = tournament;
                best_value = value;
                best_iteration = i;
                if( best_value == (w1 + w2 > 0 ? g_lower_bound : 0))
                    break;
            }
        }

        ++i;
    }

    return best_iteration;
}

void usage( const char* name)
{
    std::cout << "Usage: " << name << " <M> <N> <K> <L> <R> <I> [<V> [<w0> <w1> <w2> <w3> <w4>]]" << std::endl;
    std::cout << "    M   number of rounds, M >= 1 (*)" << std::endl;
    std::cout << "    N   number of courts, N >= 1" << std::endl;
    std::cout << "    K   number of male players (or number of total players), 2*N <= K <= 26" << std::endl;
    std::cout << "    L   number of female players, 0 or 2*N <= L <= 26" << std::endl;
    std::cout << "    R   number of runs, R >= 1 (for example, 100)" << std::endl;
    std::cout << "    I   number of iterations per run, I >= 1 (for example, 10000)" << std::endl;
    std::cout << "    V   volatility of the simulated annealing, V >= 1.0 (default: 30.0)" << std::endl;
    std::cout << "    w0  weight for multiple matches per round per player, w0 >= 0 (default: 500)" << std::endl;
    std::cout << "    w1  weight for same teammate penalty, w1 >= 0 (default: 100)" << std::endl;
    std::cout << "    w2  weight for same opponent penalty, w2 >= 0 (default: 50)" << std::endl;
    std::cout << "    w3  weight for subsequent matches, w3 >= 0 (default: 5)" << std::endl;
    std::cout << "    w4  weight for subsequent breaks, w4 >= 0 (default: 1)" << std::endl;
    std::cout << std::endl;
    std::cout << "(*) Alternatively, if this argument is prefixed by a minus sign, the argument" << std::endl;
    std::cout << "specifies the minimum number of matches per player. The number of rounds will" << std::endl;
    std::cout << "then be the smallest number to fulfill this requirement." << std::endl;
}

void ctrl_c_handler( int /*unused*/)
{
    g_interrupt = true;
}

int main( int argc, char* argv[])
{
#ifdef _WIN32
    srand( static_cast<int>( time( nullptr)));
#else // _WIN32
    srand48( static_cast<int>( time( nullptr)));
#endif // _WIN32

    // check number of arguments
    if( argc != 1+6 && argc != 1+7 && argc != 1+12) {
        usage( argv[0]);
        return 1;
    }

    // get mandatory argument
    M      = atoi( argv[1]);
    N      = atoi( argv[2]);
    K      = atoi( argv[3]);
    L      = atoi( argv[4]);
    int R  = atoi( argv[5]);
    int I  = atoi( argv[6]);

    // get optional arguments
    V      = argc >= 1+ 7 ? atof( argv[ 7]) :  30.0;
    int w0 = argc >= 1+12 ? atoi( argv[ 8]) : 500;
    int w1 = argc >= 1+12 ? atoi( argv[ 9]) : 100;
    int w2 = argc >= 1+12 ? atoi( argv[10]) :  50;
    int w3 = argc >= 1+12 ? atoi( argv[11]) :   5;
    int w4 = argc >= 1+12 ? atoi( argv[12]) :   1;

    // reject invalid arguments
    if( M == 0 || N < 1 || K < 2*N || K > 26 || L < 0 || ((0 < L) && (L < 2*N)) || L > 26 || R < 1 || I < 1 || V < 1.0 || w0 < 0 || w1 < 0 || w2 < 0 || w3 < 0 || w4 < 0) {
        usage( argv[0]);
        return 2;
    }

    // compute number of rounds if minimum number of matches per player is given
    if( M < 0) {
        int slots             = L == 0 ? -M * K : -M * std::max( K, L);
        int slots_per_court   = L == 0 ? 4 : 2;
        int slots_per_round   = N * slots_per_court;
        int rounds            = (slots + slots_per_round-1) / slots_per_round;
        int slots_last_round  = slots - (rounds-1) * slots_per_round;
        int courts_last_round = (slots_last_round + slots_per_court-1) / slots_per_court;
        M = rounds;
        last_N = courts_last_round;
    } else
        last_N = N;

    MtN = (M-1) * N + last_N;
    KpL = K + L;

    // compute lower bound for best value
    int lb_same_opponent_male   = L == 0 ? 4*MtN - K*(K-1)/2 : MtN - K*(K-1)/2;
    int lb_same_opponent_female = L == 0 ?                 0 : MtN - L*(L-1)/2;
    int lb_same_opponent_mixed  = L == 0 ?                 0 : MtN - K*L/2;
    int lb_same_teammate        = L == 0 ? 2*MtN - K*(K-1)/2 : 2*MtN - K*L;

    g_lower_bound = 0;
    if( lb_same_opponent_male > 0)
        g_lower_bound += w2 * lb_same_opponent_male;
    if( lb_same_opponent_female > 0)
        g_lower_bound += w2 * lb_same_opponent_female;
    if( lb_same_opponent_mixed > 0)
        g_lower_bound += w2 * lb_same_opponent_mixed;
    if( lb_same_teammate > 0)
        g_lower_bound += w1 * lb_same_teammate;

    Tournament best_tournament;
    int best_value = 1000000000;

    std::vector<int> best_iteration_per_run( R);

    struct sigaction sa = {};
    sa.sa_handler = ctrl_c_handler; // NOLINT cppcoreguidelines-pro-type-union-access
    sa.sa_flags = 0;
    sigemptyset( &sa.sa_mask);
    sigaction( SIGINT, &sa, nullptr);

    Timer timer;
    timer.start();

#ifdef HAVE_OPENMP
#pragma omp parallel for schedule(dynamic,16)
#endif // HAVE_OPENMP
    for( int r = 0; r < R; ++r) {

        if( g_interrupt || best_value == g_lower_bound)
            continue;

        auto buffer = std::make_unique<char[]>( KpL*KpL);

        Tournament tmp_tournament;
        // use two phases if L > 0 or in half of the runs for L == 0
        bool two_phases = L > 0 || (r % 2 == 0);
        if( two_phases) {
            // use 1/6 of the iterations to consider the matches themselves
            int I1 = I/6;
            run(  0, w1, w2,  0,  0, I1, buffer.get(), tmp_tournament);
            // use remaining iterations to arrange the matches in rounds
            int I2 = I - I1;
            best_iteration_per_run[r] = run( w0,  0,  0, w3, w4, I2, buffer.get(), tmp_tournament);
        } else
            best_iteration_per_run[r] = run( w0,  w1,  w2, w3, w4, I, buffer.get(), tmp_tournament);
        int tmp_value = evaluate( tmp_tournament, w0, w1, w2, w3, w4, buffer.get(), nullptr);

        std::ostringstream s;
        s << "run " << r << " returns value " << tmp_value
           << ", best value is " << best_value << std::endl;

#ifdef HAVE_OPENMP
#pragma omp critical
#endif // HAVE_OPENMP
        if( tmp_value < best_value) {

            best_tournament = tmp_tournament;
            best_value = tmp_value;

            s << std::endl;
            s << best_tournament;
            s << std::endl;
            evaluate( best_tournament, w0, w1, w2, w3, w4, buffer.get(), &s);
        }

        std::cerr << s.str();
    }

    timer.stop();

    // output best value
    std::cout << std::endl;
    std::cout << "best result with value " << best_value;
    if( g_lower_bound > 0)
        std::cout << ", lower bound is " << g_lower_bound;
    std::cout << std::endl << std::endl;

    // output best tournament
    std::cout << best_tournament;
    std::cout << std::endl;

    // output penalties
    auto buffer = std::make_unique<char[]>( KpL*KpL);
    evaluate( best_tournament, w0, w1, w2, w3, w4, buffer.get(), &std::cout);

    // output rounds
    std::vector<std::list<int> > rounds;
    rounds.resize( KpL);
    for( int i = 0; i < best_tournament.get_rounds(); ++i) {
        for( int j = 0; j < best_tournament.get_courts( i); ++j) {
            const Match& match = best_tournament.get_match( i, j);
            for( int k = 0; k < 4; ++k)
                rounds[match.get_player( k).get_id()].push_back( i);
        }
    }
    for( int p = 0; p < KpL; ++ p) {
        auto it = rounds[p].begin();
        if( it == rounds[p].end()) {
           std::cout << "player " << Player( p) << " plays never" << std::endl;
        } else {
           std::cout << "player " << Player( p) << " plays in round(s) " << (*it++)+1;
           for( ; it != rounds[p].end(); ++it)
               std::cout << ", " << (*it) + 1;
           std::cout << " (" << rounds[p].size()
                     << (rounds[p].size() == 1 ? " match)" : " matches)") << std::endl;
        }
    }
    std::cout << std::endl;

    // output mean and standard deviation of best iteration
    double mean = std::accumulate( best_iteration_per_run.begin(), best_iteration_per_run.end(), 0.0) / R;
    double t = 0;
    for( int i = 0; i < R; ++i)
        t += (best_iteration_per_run[i]-mean) * (best_iteration_per_run[i]-mean);
    double std_dev = sqrt( t / R);
    std::cerr << "mean: " << mean << ", std. deviation: " << std_dev << std::endl;

    std::cerr << std::endl;
    std::cerr << "time: " << timer.get_real_time() << " seconds" << std::endl;

    sa.sa_handler = SIG_DFL; // NOLINT cppcoreguidelines-pro-type-union-access
    sa.sa_flags = 0;
    sigemptyset( &sa.sa_mask);
    sigaction( SIGINT, &sa, nullptr);

    return 0;
}
